


/*
 *  $Id: chsend.c,v 1.18 1994/07/13 15:23:12 gropp Exp gropp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char SCCSid[] = "%W% %G%";
#endif

#include "mpid.h"

/* 
   Still need to do:  clean up the post_short, post_long to look more like the
   code in chrecv.c .  Complicated slightly because the chrecv has already
   received part of the message, where here, the message header needs to
   be sent with, possibly, some of the data.

   There are many other strategies for IMPLEMENTING the ADI than the one
   shown here.  For example, a more deliberate packetizing strategy could 
   be used.  For systems with interrupt-driven receives, we could send 
   messages only in responce to a request.  If access to lower levels of
   the transport is available, then the protocols for transmitting a message
   can be customized to the ADI.
 */

/***************************************************************************/
/* These routines enable the debugging output                              */
/***************************************************************************/
static int DebugFlag = 0;   /* Set to 0 for no debug output */

void MPID_SetSendDebugFlag( f )
int f;
{
DebugFlag = f;
}
/***************************************************************************/

/* 
   This file includes the routines to handle the device part of a send
   for Chameleon

   As a reminder, the first element is the device handle, the second is
   the (basically opaque) mpi handle
 */

/* Send a short (single packet message) */
int MPID_NX_post_send_short( dmpi_send_handle, len ) 
MPIR_SHANDLE *dmpi_send_handle;
int len;
{
MPID_SHANDLE *mpid_send_handle;
char        *address;
MPID_PACKET pkt;
int         actual_len;
int         dest;

/* These references are ordered to match the order they appear in the 
   structure */
mpid_send_handle = &dmpi_send_handle->dev_shandle;
dest             = dmpi_send_handle->dest;
pkt.tag		 = dmpi_send_handle->tag;
pkt.context_id	 = dmpi_send_handle->contextid;
pkt.mode	 = dmpi_send_handle->mode;
#ifdef MPID_HAS_HETERO
if (dmpi_send_handle->msgrep == MPIR_MSGREP_XDR) 
    pkt.mode |= MPID_MODE_XDR;
#endif
pkt.lrank	 = dmpi_send_handle->lrank;
MPID_MSG_SET_LEN(&pkt,len);
if (MPID_MODE_IS_SYNC(&pkt))
    MPID_NX_Get_Sync_Id( &pkt, dmpi_send_handle, mpid_send_handle );

if (DebugFlag) {
    printf( 
 "[%d]S Starting a send of tag = %d, len = %d, ctx = %d, dest = %d, mode=%x\n",
	    __MYPROCID, pkt.tag, pkt.len, pkt.context_id, dest, pkt.mode );
    fflush( stdout );
    }

#ifdef MPID_HAS_HETERO
/* Convert the header into canonical integer order */
if (MPID_IS_HETERO &&
    MPID_procinfo[dest].byte_order != MPID_byte_order) {
    /* Need to swap to receiver's order.  We ALWAYS reorder at the
       sender's end (this is because a message can be received with MPI_Recv
       instead of MPI_Recv/MPI_Unpack, and hence requires us to use a format
       that matches the receiver's ordering without requiring a user-unpack.
       We may need to generalize this for XDR systems... */
    SYByteSwapInt( &pkt, MPID_HEADER_INTS );
    }
#endif

if (mpid_send_handle->start) {
    memcpy( pkt.buffer, mpid_send_handle->start, len );
    if (DebugFlag) {
	MPID_NX_Print_pkt_data( "S Getting data from mpid->start", 
			        pkt.buffer, len );
	}
    }
else if (len > 0) {
    /* Note that if len is 0, we may not have a ->start pointer but we
       still don't need to get any data */
    actual_len = len;
    DMPI_get_into_contig( dmpi_send_handle, (void *)(pkt.buffer), 
			  MPID_PACKET_SIZE, &actual_len );
    if (DebugFlag) {
	printf( 
      "[%d]S Getting data (%d bytes) with get_into_contig, first int is %d\n",
	        __MYPROCID, actual_len, *(int *)&pkt.buffer[0] );
	MPID_NX_Print_pkt_data( (char *)0, pkt.buffer, pkt.len );
	}
    }
/* Always use a blocking send for short messages.
   (May fail with systems that do not provide adequate
   buffering.  These systems should switch to non-blocking sends)
 */
if (DebugFlag) {
	printf( "[%d]S Sending message in a single packet...\n", __MYPROCID );
	fflush( stdout );
	}
/* In case the message is marked as non-blocking, indicate that we don't
   need to wait on it */
mpid_send_handle->sid = 0;
_csend(MPID_PT2PT_TAG,&pkt,len + MPID_HEADER_LEN,dest,0);
if (!MPID_MODE_IS_SYNC(&pkt)) {
    mpid_send_handle->done = MPIR_YES;
    DMPI_mark_send_completed( dmpi_send_handle );
    }
if (DebugFlag) {
    printf( "[%d]S Sent message in a single packet...\n", __MYPROCID );
    fflush( stdout );
    }
return MPI_SUCCESS;
}

/* Long message */
int MPID_NX_post_send_long( dmpi_send_handle, len ) 
MPIR_SHANDLE *dmpi_send_handle;
int len;
{
MPID_SHANDLE     *mpid_send_handle;
char             *address;
MPID_PACKET      pkt;
int              actual_len;
int              dest;

mpid_send_handle = &dmpi_send_handle->dev_shandle;
pkt.len	       = len;
pkt.context_id = dmpi_send_handle->contextid;
pkt.tag	       = dmpi_send_handle->tag;
/* Modes should be set with either the devices types or by converting from
   MPIR to the ADI's prefered modes */
pkt.mode       = dmpi_send_handle->mode;
#ifdef MPID_HAS_HETERO
if (dmpi_send_handle->msgrep == MPIR_MSGREP_XDR) 
    pkt.mode |= MPID_MODE_XDR;
#endif
pkt.lrank      = dmpi_send_handle->lrank;
dest           = dmpi_send_handle->dest;
if (MPID_MODE_IS_SYNC(&pkt))
    MPID_NX_Get_Sync_Id( &pkt, dmpi_send_handle, mpid_send_handle );


if (DebugFlag) {
    printf( 
 "[%d]S Starting a send of tag = %d, len = %d, ctx = %d, dest = %d, mode=%x\n",
	    __MYPROCID, pkt.tag, pkt.len, pkt.context_id, dest, pkt.mode );
    fflush( stdout );
    }

#ifdef MPID_HAS_HETERO
/* Convert the header into canonical integer order */
if (MPID_IS_HETERO &&
    MPID_procinfo[dest].byte_order != MPID_byte_order) {
    /* Need to swap to receivers order.  We ALWAYS reorder at the
       sender's end */
    SYByteSwapInt( &pkt, MPID_HEADER_INTS );
    }
#endif

if (DebugFlag) {
    if (mpid_send_handle->start) {
	printf( "[%d]S Getting data from mpid->start, first int is %d\n",
	        __MYPROCID, *(int *)&pkt.buffer[0] );
	fflush( stdout );
	}
    else if (len > 0) {
	/* Note that if len is 0, we may not have a ->start pointer but we
	   still don't need to get any data */
	printf( 
       "[%d]S Getting data (%d bytes) with get_into_contig, first int is %d\n",
	        __MYPROCID, len, *(int *)&pkt.buffer[0] );
	fflush( stdout );
	}
    }
if (DebugFlag) {
    printf( "[%d]S Sending extra-long message...\n", __MYPROCID );
    fflush( stdout );
    }
/* Send as packet only */
_csend(MPID_PT2PT_TAG,&pkt,MPID_HEADER_LEN,dest,0);
mpid_send_handle->sid = 0;
do {
    actual_len = len;
    if (mpid_send_handle->start) {
	address    = ((char*)mpid_send_handle->start);
	}
    else {
	DMPI_get_from_contig( dmpi_send_handle, (void **)(&address), 
			     -1, &actual_len );
	}
#ifndef PI_NO_NSEND
    if (mpid_send_handle->is_non_blocking) {
	if (mpid_send_handle->sid) {
	    /* Assumes that only the sid field
	       is used; this is ok as long as the PI_NO_NSEND 
	       wrappers are used... */
	    msgwait(mpid_send_handle->sid );
	    }
	mpid_send_handle->sid =_isend(MPID_PT2PT2_TAG(__MYPROCID),address,actual_len,dest,0);
	}
    else 
#endif
	{
	_csend(MPID_PT2PT2_TAG(__MYPROCID),address,actual_len,dest,0);
	}
    len -= actual_len;
    } while (len > 0);
#ifndef PI_NO_NSEND
if (!MPID_MODE_IS_SYNC(&pkt)) {
    mpid_send_handle->done = 1;
    DMPI_mark_send_completed( dmpi_send_handle );
    }
#else
if (dmpi_send_handle->mode != MPIR_MODE_SYNCHRONOUS) {
    mpid_send_handle->done = MPIR_YES;
    DMPI_mark_send_completed( dmpi_send_handle );
    }
#endif
return MPI_SUCCESS;
}


/*
   This implementation immediately sends the data.

   It takes advantage of being provided with the address of the user-buffer
   in the contiguous case.
 */
int MPID_NX_post_send( dmpi_send_handle ) 
MPIR_SHANDLE *dmpi_send_handle;
{
MPID_SHANDLE *mpid_send_handle;
char        *address;
MPID_PACKET pkt;
int         actual_len, len;
int         dest;

mpid_send_handle = &dmpi_send_handle->dev_shandle;
/* Get the header */
if (mpid_send_handle->start) {
    actual_len = mpid_send_handle->bytes_as_contig;
    }
else {
    DMPI_get_totallen( dmpi_send_handle, &actual_len );
    }
if (actual_len > MPID_PACKET_SIZE) 
    return MPID_NX_post_send_long( dmpi_send_handle, actual_len );
else
    return MPID_NX_post_send_short( dmpi_send_handle, actual_len );
}


/*
  Chameleon gets no asynchronous notice that the message has been complete,
  so there is no asynchronous ref to DMPI_mark_send_completed.
 */
int MPID_NX_isend_wait( dmpi_send_handle )
MPIR_SHANDLE *dmpi_send_handle;
{
MPID_SHANDLE *mpid_send_handle;

mpid_send_handle = &dmpi_send_handle->dev_shandle;

/* Wait on the message */
#ifndef PI_NO_NSEND
if (mpid_send_handle->sid) {
    /* We don't use non-blocking if the message is short enough... */
    msgwait(mpid_send_handle->sid );
    mpid_send_handle->sid = 0;
    }
#endif
if (dmpi_send_handle->mode != MPIR_MODE_SYNCHRONOUS) {
    mpid_send_handle->done = MPIR_YES;
    DMPI_mark_send_completed( dmpi_send_handle );
    }
/* CHECK - if a nonblocking, synchronous send is used, we may wait on it
   again when complete send is called.  We must make sure that in this case, 
   only MPID_NX_complete_send calls this routine 
 */

return MPI_SUCCESS;
}

int MPID_NX_complete_send( dmpi_send_handle, status ) 
MPIR_SHANDLE *dmpi_send_handle;
MPI_Status   *status;
{
MPID_SHANDLE *mpid_send_handle = &dmpi_send_handle->dev_shandle;

/* Check to see if we need to complete the send. */
if (DebugFlag) {
	printf( "[%d]S Entering complete send...\n", __MYPROCID );
	fflush( stdout );
	}
if (mpid_send_handle->sid)  {
    MPID_NX_isend_wait( dmpi_send_handle );
    }
if (DebugFlag) {
	printf( "[%d]S Entering complete send while loop...\n", __MYPROCID );
	fflush( stdout );
	}
while (!mpid_send_handle->done) {
    /* This waits for the completion of a synchronous send, since at
       this point, we've finished waiting for the =_isend(,,,,0) to complete */
    (void)MPID_NX_check_incoming( MPID_BLOCKING );
    }
if (DebugFlag) {
	printf( "[%d]S Exiting complete send...\n", __MYPROCID );
	fflush( stdout );
	}
}





