/*
 *  $Id: chsend.c,v 1.21 1994/09/11 15:25:07 gropp Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      All rights reserved.  See COPYRIGHT in top-level directory.
 */

#ifndef lint
static char SCCSid[] = "%W% %G%";
#endif

#include "mpid.h"

/* 
   Still need to do:  clean up the post_short, post_long to look more like the
   code in chrecv.c .  Complicated slightly because the chrecv has already
   received part of the message, where here, the message header needs to
   be sent with, possibly, some of the data.

   There are many other strategies for IMPLEMENTING the ADI than the one
   shown here.  For example, a more deliberate packetizing strategy could 
   be used.  For systems with interrupt-driven receives, we could send 
   messages only in responce to a request.  If access to lower levels of
   the transport is available, then the protocols for transmitting a message
   can be customized to the ADI.
 */

/***************************************************************************/
/* These routines enable the debugging output                              */
/***************************************************************************/
static int DebugFlag = 0;   /* Set to 0 for no debug output */

void MPID_SetSendDebugFlag( f )
int f;
{
DebugFlag = f;
}
/***************************************************************************/

/* 
   This file includes the routines to handle the device part of a send
   for Chameleon

   As a reminder, the first element is the device handle, the second is
   the (basically opaque) mpi handle
 */

/* Send a short (single packet message) */
int MPID_CH_post_send_short( dmpi_send_handle, len ) 
MPIR_SHANDLE *dmpi_send_handle;
int len;
{
MPID_SHANDLE *mpid_send_handle;
char        *address;
MPID_PACKET pkt;
int         actual_len;
int         dest;

/* These references are ordered to match the order they appear in the 
   structure */
mpid_send_handle = &dmpi_send_handle->dev_shandle;
dest             = dmpi_send_handle->dest;
pkt.tag		 = dmpi_send_handle->tag;
pkt.context_id	 = dmpi_send_handle->contextid;
pkt.mode	 = dmpi_send_handle->mode;
#ifdef MPID_HAS_HETERO         /* #HETERO_START# */
if (dmpi_send_handle->msgrep == MPIR_MSGREP_XDR) 
    pkt.mode |= MPID_MODE_XDR;
#endif                         /* #HETERO_END# */
pkt.lrank	 = dmpi_send_handle->lrank;
MPID_MSG_SET_LEN(&pkt,len);
if (MPID_MODE_IS_SYNC(&pkt))
    MPID_CH_Get_Sync_Id( &pkt, dmpi_send_handle, mpid_send_handle );

#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
if (DebugFlag) {
    printf( 
 "[%d]S Starting a send of tag = %d, len = %d, ctx = %d, dest = %d, mode=",
	    PImytid, pkt.tag, pkt.len, pkt.context_id, dest );
    MPID_Print_mode( stdout, &pkt );
    fprintf( stdout, "(%s:%d)\n", __FILE__, __LINE__ );
    fflush( stdout );
    }
#endif                  /* #DEBUG_END# */

#ifdef MPID_HAS_HETERO       /* #HETERO_START# */
/* Convert the header into canonical integer order */
if (MPID_IS_HETERO &&
    MPID_procinfo[dest].byte_order != MPID_byte_order) {
    /* Need to swap to receiver's order.  We ALWAYS reorder at the
       sender's end (this is because a message can be received with MPI_Recv
       instead of MPI_Recv/MPI_Unpack, and hence requires us to use a format
       that matches the receiver's ordering without requiring a user-unpack.
       We may need to generalize this for XDR systems... */
    SYByteSwapInt( &pkt, MPID_HEADER_INTS );
    }
#endif                       /* #HETERO_END# */

if (mpid_send_handle->start) {
    memcpy( pkt.buffer, mpid_send_handle->start, len );
#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
    if (DebugFlag) {
	MPID_CH_Print_pkt_data( "S Getting data from mpid->start", 
			        pkt.buffer, len );
	}
#endif                  /* #DEBUG_END# */
    }
else if (len > 0) {
    /* Note that if len is 0, we may not have a ->start pointer but we
       still don't need to get any data */
    actual_len = len;
    DMPI_get_into_contig( dmpi_send_handle, (void *)(pkt.buffer), 
			  MPID_PACKET_SIZE, &actual_len );
#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
    if (DebugFlag) {
	printf( 
"[%d]S Getting data (%d bytes) with get_into_contig, first int is %d (%s:%d)\n",
	        PImytid, actual_len, *(int *)&pkt.buffer[0], 
	        __FILE__, __LINE__ );
	MPID_CH_Print_pkt_data( (char *)0, pkt.buffer, pkt.len );
	}
#endif                  /* #DEBUG_END# */
    }
/* Always use a blocking send for short messages.
   (May fail with systems that do not provide adequate
   buffering.  These systems should switch to non-blocking sends)
 */
#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
if (DebugFlag) {
	printf( "[%d]S Sending message in a single packet (%s:%d)...\n", 
	       PImytid, __FILE__, __LINE__ );
	MPID_Print_packet( stdout, &pkt );
	fflush( stdout );
	}
#endif                  /* #DEBUG_END# */
/* In case the message is marked as non-blocking, indicate that we don't
   need to wait on it */
mpid_send_handle->sid = 0;
PIbsend( MPID_PT2PT_TAG, &pkt, len + MPID_HEADER_LEN, dest, MSG_OTHER );
if (!MPID_MODE_IS_SYNC(&pkt)) {
    mpid_send_handle->done = MPIR_YES;
    DMPI_mark_send_completed( dmpi_send_handle );
    }
#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
if (DebugFlag) {
    printf( "[%d]S Sent message in a single packet (%s:%d)...\n", 
	    PImytid, __FILE__, __LINE__ );
    fflush( stdout );
    }
#endif                  /* #DEBUG_END# */
return MPI_SUCCESS;
}

/* Long message */
int MPID_CH_post_send_long( dmpi_send_handle, len ) 
MPIR_SHANDLE *dmpi_send_handle;
int len;
{
MPID_SHANDLE     *mpid_send_handle;
char             *address;
MPID_PACKET      pkt;
int              actual_len;
int              dest;

mpid_send_handle = &dmpi_send_handle->dev_shandle;
pkt.len	       = len;
pkt.context_id = dmpi_send_handle->contextid;
pkt.tag	       = dmpi_send_handle->tag;
/* Modes should be set with either the devices types or by converting from
   MPIR to the ADI's prefered modes */
pkt.mode       = dmpi_send_handle->mode;
#ifdef MPID_HAS_HETERO      /* #HETERO_START# */
if (dmpi_send_handle->msgrep == MPIR_MSGREP_XDR) 
    pkt.mode |= MPID_MODE_XDR;
#endif                      /* #HETERO_END# */
pkt.lrank      = dmpi_send_handle->lrank;
dest           = dmpi_send_handle->dest;
if (MPID_MODE_IS_SYNC(&pkt))
    MPID_CH_Get_Sync_Id( &pkt, dmpi_send_handle, mpid_send_handle );


#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
if (DebugFlag) {
    printf( 
 "[%d]S Starting a send of tag = %d, len = %d, ctx = %d, dest = %d, mode=",
	    PImytid, pkt.tag, pkt.len, pkt.context_id, dest );
    MPID_Print_mode( stdout, &pkt );
    fprintf( stdout, "(%s:%d)\n", __FILE__, __LINE__ );
    fflush( stdout );
    }
#endif                  /* #DEBUG_END# */

#ifdef MPID_HAS_HETERO         /* #HETERO_START# */
/* Convert the header into canonical integer order */
if (MPID_IS_HETERO &&
    MPID_procinfo[dest].byte_order != MPID_byte_order) {
    /* Need to swap to receivers order.  We ALWAYS reorder at the
       sender's end */
    SYByteSwapInt( &pkt, MPID_HEADER_INTS );
    }
#endif                         /* #HETERO_END# */

#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
if (DebugFlag) {
    if (mpid_send_handle->start) {
	printf( 
	   "[%d]S Getting data from mpid->start, first int is %d (%s:%d)\n",
	        PImytid, *(int *)&pkt.buffer[0], __FILE__, __LINE__ );
	fflush( stdout );
	}
    else if (len > 0) {
	/* Note that if len is 0, we may not have a ->start pointer but we
	   still don't need to get any data */
	printf( 
"[%d]S Getting data (%d bytes) with get_into_contig, first int is %d (%s:%d)\n",
	        PImytid, len, *(int *)&pkt.buffer[0], __FILE__, __LINE__ );
	fflush( stdout );
	}
    printf( "[%d]S Sending extra-long message (%s:%d)...\n", 
	    PImytid, __FILE__, __LINE__ );
    MPID_Print_packet( stdout, &pkt );
    fflush( stdout );
    }
#endif                  /* #DEBUG_END# */
/* Send as packet only */
PIbsend( MPID_PT2PT_TAG, &pkt, MPID_HEADER_LEN, dest, MSG_OTHER );
mpid_send_handle->sid = 0;
do {
    actual_len = len;
    if (mpid_send_handle->start) {
	address    = ((char*)mpid_send_handle->start);
	}
    else {
	DMPI_get_from_contig( dmpi_send_handle, (void **)(&address), 
			     -1, &actual_len );
	}
#ifndef PI_NO_NSEND
    if (mpid_send_handle->is_non_blocking) {
	if (mpid_send_handle->sid) {
	    /* Assumes that only the sid field
	       is used; this is ok as long as the PI_NO_NSEND 
	       wrappers are used... */
	    PIwsend( MPID_PT2PT2_TAG(PImytid), address, actual_len, 0, 
		    MSG_OTHER, mpid_send_handle->sid );
	    }
	PInsend( MPID_PT2PT2_TAG(PImytid), address, actual_len, dest, 
		MSG_OTHER, mpid_send_handle->sid );
	}
    else 
#endif
	{
	PIbsend( MPID_PT2PT2_TAG(PImytid), address, actual_len, dest, 
		MSG_OTHER );
	}
    len -= actual_len;
    } while (len > 0);
#ifndef PI_NO_NSEND
if (!MPID_MODE_IS_SYNC(&pkt)) {
    mpid_send_handle->done = 1;
    DMPI_mark_send_completed( dmpi_send_handle );
    }
#else
if (dmpi_send_handle->mode != MPIR_MODE_SYNCHRONOUS) {
    mpid_send_handle->done = MPIR_YES;
    DMPI_mark_send_completed( dmpi_send_handle );
    }
#endif
return MPI_SUCCESS;
}


/*
   This implementation immediately sends the data.

   It takes advantage of being provided with the address of the user-buffer
   in the contiguous case.
 */
int MPID_CH_post_send( dmpi_send_handle ) 
MPIR_SHANDLE *dmpi_send_handle;
{
MPID_SHANDLE *mpid_send_handle;
char        *address;
MPID_PACKET pkt;
int         actual_len, len;
int         dest;

mpid_send_handle = &dmpi_send_handle->dev_shandle;
/* Get the header */
if (mpid_send_handle->start) {
    actual_len = mpid_send_handle->bytes_as_contig;
    }
else {
    DMPI_get_totallen( dmpi_send_handle, &actual_len );
    }
if (actual_len > MPID_PACKET_SIZE) 
    return MPID_CH_post_send_long( dmpi_send_handle, actual_len );
else
    return MPID_CH_post_send_short( dmpi_send_handle, actual_len );
}


/*
  Chameleon gets no asynchronous notice that the message has been complete,
  so there is no asynchronous ref to DMPI_mark_send_completed.
 */
int MPID_CH_isend_wait( dmpi_send_handle )
MPIR_SHANDLE *dmpi_send_handle;
{
MPID_SHANDLE *mpid_send_handle;

mpid_send_handle = &dmpi_send_handle->dev_shandle;

/* Wait on the message */
#ifndef PI_NO_NSEND
if (mpid_send_handle->sid) {
    /* We don't use non-blocking if the message is short enough... */
    PIwsend( MPID_PT2PTTAG(mpid_send_handle->to), (void *)0,
	    mpid_send_handle->len, mpid_send_handle->to, MSG_OTHER, 
	    mpid_send_handle->sid );
    mpid_send_handle->sid = 0;
    }
#endif
if (dmpi_send_handle->mode != MPIR_MODE_SYNCHRONOUS) {
    mpid_send_handle->done = MPIR_YES;
    DMPI_mark_send_completed( dmpi_send_handle );
    }
/* CHECK - if a nonblocking, synchronous send is used, we may wait on it
   again when complete send is called.  We must make sure that in this case, 
   only MPID_CH_complete_send calls this routine 
 */

return MPI_SUCCESS;
}

/*
  We have to be careful here.  If the wait would block because a matching
  receive has not yet been posted on the destination end, we could deadlock.

  The "solution" here is to first clear any incoming messages.  This allows
  us to post a matching receive that this send is supposed to complete.
  This solution is not complete; there are race conditions that can still
  cause it to fail.  In addition, the current code to handle incoming messages
  may try to force the receive to complete first; this will cause some systems
  to deadlock.  We probably need to packetize to guarentee reliable 
  behavior, and allow for partial completion.  

  Deferred to a later implementation.
 */
int MPID_CH_complete_send( dmpi_send_handle, status ) 
MPIR_SHANDLE *dmpi_send_handle;
MPI_Status   *status;
{
MPID_SHANDLE *mpid_send_handle = &dmpi_send_handle->dev_shandle;

/* Check to see if we need to complete the send. */
#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
if (DebugFlag) {
	printf( "[%d]S Entering complete send (%s:%d)...\n", 
	        PImytid, __FILE__, __LINE__ );
	fflush( stdout );
	}
#endif                  /* #DEBUG_END# */
if (mpid_send_handle->sid)  {
    /* Before we do the wait, try to clear all pending messages */
    (void)MPID_CH_check_incoming( MPID_NOTBLOCKING );
    MPID_CH_isend_wait( dmpi_send_handle );
    }
#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
if (DebugFlag) {
	printf( "[%d]S Entering complete send while loop (%s:%d)...\n", 
	        PImytid, __FILE__, __LINE__ );
	fflush( stdout );
	}
#endif                  /* #DEBUG_END# */
while (!mpid_send_handle->done) {
    /* This waits for the completion of a synchronous send, since at
       this point, we've finished waiting for the PInsend to complete */
    (void)MPID_CH_check_incoming( MPID_BLOCKING );
    }
#ifdef MPID_DEBUG_ALL   /* #DEBUG_START# */
if (DebugFlag) {
    printf( "[%d]S Exiting complete send (%s:%d)...\n", 
	   PImytid, __FILE__, __LINE__ );
    fflush( stdout );
    }
#endif                  /* #DEBUG_END# */
}

int MPID_Print_mode( fp, pkt )
FILE        *fp;
MPID_PACKET *pkt;
{
char *modename=0;
if (MPID_MODE_HAS_XDR(pkt)) fputs( "xdr", fp );
if (MPID_MODE_IS_READY(pkt)) fputs( "ready", fp );

if (MPID_MODE_IS_SYNCACK(pkt)) modename = "syncack";
else if (MPID_MODE_IS_SYNC(pkt)) modename = "sync";
if (modename)
    fputs( modename, fp );
if (MPID_MODE_IS_SYNC(pkt) || MPID_MODE_IS_SYNCACK(pkt)) {
    fprintf( fp, " - id = %d", MPID_SYNC_GET_ID(pkt) );
    }
}

int MPID_Print_packet( fp, pkt )
FILE        *fp;
MPID_PACKET *pkt;
{
fprintf( fp, "[%d] PKT = \n\
\tlen        = %d\n\
\ttag        = %d\n\
\tcontent_id = %d\n\
\tlrank      = %d\n\
\tmode       = ", 
	PImytid, pkt->len, pkt->tag, pkt->context_id, pkt->lrank );
MPID_Print_mode( fp, pkt );
fputs( "\n", fp );
}


