/*
 *  $Id: recv.c,v 1.18 1994/08/04 20:55:59 bridges Exp $
 *
 *  (C) 1993 by Argonne National Laboratory and Mississipi State University.
 *      See COPYRIGHT in top-level directory.
 */


#ifndef lint
static char vcid[] = "$Id: recv.c,v 1.18 1994/08/04 20:55:59 bridges Exp $";
#endif /* lint */

#include "mpiimpl.h"
/*@
    MPI_Recv - Basic receive

Output Parameter:
. buf - initial address of receive buffer (choice) 
. status - status object (Status) 

Input Parameters:
. count - number of elements in receive buffer (integer) 
. datatype - datatype of each receive buffer element (handle) 
. source - rank of source (integer) 
. tag - message tag (integer) 
. comm - communicator (handle) 

@*/
int MPI_Recv( buf, count, datatype, source, tag, comm, status )
void             *buf;
int              count, source, tag;
MPI_Datatype     datatype;
MPI_Comm         comm;
MPI_Status       *status;
{
    MPI_Request request;
    MPIR_RHANDLE rhandle;
    int         errno, is_available;

    /* 
       Because this is a very common routine, we show how it can be
       optimized to be run "inline"; In addition, this lets us exploit
       features in the ADI to simplify the execution of blocking receive
       calls.
     */
    if (source != MPI_PROC_NULL)
    {
        if (MPIR_TEST_COMM(comm,comm) || MPIR_TEST_COUNT(comm,count) ||
	    MPIR_TEST_DATATYPE(comm,datatype) || 
	    MPIR_TEST_RECV_TAG(comm,tag) ||
	    MPIR_TEST_RECV_RANK(comm,source)) 
	    return MPIR_ERROR(comm, errno, "Error in MPI_RECV" );

        request = (MPI_Request)&rhandle;
	request->type	    = MPIR_RECV;
	if (source == MPI_ANY_SOURCE)
	    rhandle.source  = source;
	else
	    rhandle.source  = comm->group->lrank_to_grank[source];
	rhandle.tag	    = tag;
	rhandle.contextid   = comm->recv_context;
	rhandle.comm        = comm;
	rhandle.datatype    = datatype;
	rhandle.bufadd	    = buf;
	rhandle.count	    = count;
	rhandle.completed   = MPIR_NO;
#ifdef MPID_HAS_HETERO
	rhandle.msgrep      = MPIR_MSGREP_UNKNOWN;
#endif
	MPID_alloc_recv_handle(&(rhandle.dev_rhandle));

	MPIR_Receive_setup(&request);

#ifdef MPID_blocking_recv
	/* If ADI supports blocking receive directly, use that */
	MPID_blocking_recv( &rhandle );
#else
        /* ADI will handle queueing of MPIR recv handle in posted-recv queue */
	MPID_set_recv_is_nonblocking( &(rhandle.dev_rhandle), 1 );
	MPID_post_recv( &rhandle, &(is_available) );
	while ( request->rhandle.completed == MPIR_NO ) {
	    MPID_check_device( MPID_BLOCKING );
	    }
	MPID_complete_recv( &rhandle );
#endif
	status->MPI_SOURCE     = rhandle.source;
	status->MPI_TAG        = rhandle.tag;
	status->count	       = rhandle.actcount;  
	status->count_in_bytes = rhandle.totallen;

#ifdef MPID_RETURN_PACKED
	if (rhandle.bufpos) 
	    errno = MPIR_UnPackMessage( buf, count, datatype, 
				        source, request );
#endif
    }
    else {
	/* See MPI standard section 3.11 */
	status->MPI_SOURCE     = MPI_PROC_NULL;
	status->MPI_TAG        = MPI_ANY_TAG;
	status->count	       = 0;
	status->count_in_bytes = 0;
	}
    return MPI_SUCCESS;
}

